@@ -1,5 +1,6 @@
 # -*- coding: utf-8 -*-
 import time
+from datetime import datetime
 
 try:
     import ujson as json
@@ -28,6 +29,7 @@ FILE_URL = None
 
 CONNECT_TIMEOUT = 3.5
 READ_TIMEOUT = 9999
+SESSION_TIME_TO_LIVE = None  # In seconds. None - live forever, 0 - one-time
 
 RETRY_ON_ERROR = False
 RETRY_TIMEOUT = 2
@@ -39,7 +41,21 @@ ENABLE_MIDDLEWARE = False
 
 
 def _get_req_session(reset=False):
-    return util.per_thread('req_session', lambda: session if session else requests.session(), reset)
+    if SESSION_TIME_TO_LIVE:
+        # If session TTL is set - check time passed
+        creation_date = util.per_thread('req_session_time', lambda: datetime.now(), reset)
+        if (datetime.now() - creation_date).total_seconds() > SESSION_TIME_TO_LIVE:
+            # Force session reset
+            reset = True
+            # Save reset time
+            util.per_thread('req_session_time', lambda: datetime.now(), True)
+
+    if SESSION_TIME_TO_LIVE == 0:
+        # Session is one-time use
+        return requests.sessions.Session()
+    else:
+        # Session lives some time or forever once created. Default
+        return util.per_thread('req_session', lambda: session if session else requests.sessions.Session(), reset)
 
 
 def _make_request(token, method_name, method='get', params=None, files=None):
@@ -57,7 +73,7 @@ def _make_request(token, method_name, method='get', params=None, files=None):
     else:
         request_url = "https://api.telegram.org/bot{0}/{1}".format(token, method_name)
 
-    logger.debug("Request: method={0} url={1} params={2} files={3}".format(method, request_url, params, files))
+    logger.debug("Request: method={0} url={1} params={2} files={3}".format(method, request_url, params, files).replace(token, token.split(':')[0] + ":{TOKEN}"))
     read_timeout = READ_TIMEOUT
     connect_timeout = CONNECT_TIMEOUT
     if files and format_header_param:
@@ -68,8 +84,11 @@ def _make_request(token, method_name, method='get', params=None, files=None):
         if 'connect-timeout' in params:
             connect_timeout = params.pop('connect-timeout') + 10
         if 'long_polling_timeout' in params:
-            # For getUpdates: the only function with timeout on the BOT API side
+            # For getUpdates
+            # The only function with timeout on the BOT API side
             params['timeout'] = params.pop('long_polling_timeout')
+            # Long polling hangs for given time. Read timeout should be greater that long_polling_timeout
+            read_timeout = max(params['timeout'] + 10, read_timeout)
 
 
     result = None
@@ -206,7 +225,8 @@ def send_message(
     return _make_request(token, method_url, params=payload, method='post')
 
 
-def set_webhook(token, url=None, certificate=None, max_connections=None, allowed_updates=None, ip_address=None, timeout=None):
+def set_webhook(token, url=None, certificate=None, max_connections=None, allowed_updates=None, ip_address=None,
+                drop_pending_updates = None, timeout=None):
     method_url = r'setWebhook'
     payload = {
         'url': url if url else "",
@@ -216,10 +236,12 @@ def set_webhook(token, url=None, certificate=None, max_connections=None, allowed
         files = {'certificate': certificate}
     if max_connections:
         payload['max_connections'] = max_connections
-    if allowed_updates is not None:  # Empty lists should pass
+    if allowed_updates is not None:       # Empty lists should pass
         payload['allowed_updates'] = json.dumps(allowed_updates)
-    if ip_address is not None:       # Empty string should pass
+    if ip_address is not None:            # Empty string should pass
         payload['ip_address'] = ip_address
+    if drop_pending_updates is not None:  # Any bool value should pass
+        payload['drop_pending_updates'] = drop_pending_updates
     if timeout:
         payload['connect-timeout'] = timeout
     return _make_request(token, method_url, params=payload, files=files)
@@ -228,7 +250,7 @@ def set_webhook(token, url=None, certificate=None, max_connections=None, allowed
 def delete_webhook(token, drop_pending_updates=None, timeout=None):
     method_url = r'deleteWebhook'
     payload = {}
-    if drop_pending_updates is not None:  # None / True / False
+    if drop_pending_updates is not None:  # Any bool value should pass
         payload['drop_pending_updates'] = drop_pending_updates
     if timeout:
         payload['connect-timeout'] = timeout
@@ -251,7 +273,7 @@ def get_updates(token, offset=None, limit=None, timeout=None, allowed_updates=No
     if limit:
         payload['limit'] = limit
     if timeout:
-        payload['timeout'] = timeout
+        payload['connect-timeout'] = timeout
     if long_polling_timeout:
         payload['long_polling_timeout'] = long_polling_timeout
     if allowed_updates is not None:  # Empty lists should pass
@@ -323,6 +345,30 @@ def forward_message(
     return _make_request(token, method_url, params=payload)
 
 
+def copy_message(token, chat_id, from_chat_id, message_id, caption=None, parse_mode=None, caption_entities=None,
+                 disable_notification=None, reply_to_message_id=None, allow_sending_without_reply=None,
+                 reply_markup=None, timeout=None):
+    method_url = r'copyMessage'
+    payload = {'chat_id': chat_id, 'from_chat_id': from_chat_id, 'message_id': message_id}
+    if caption is not None:
+        payload['caption'] = caption
+    if parse_mode:
+        payload['parse_mode'] = parse_mode
+    if caption_entities is not None:
+        payload['caption_entities'] = _convert_entites(caption_entities)
+    if disable_notification is not None:
+        payload['disable_notification'] = disable_notification
+    if reply_to_message_id:
+        payload['reply_to_message_id'] = reply_to_message_id
+    if reply_markup is not None:
+        payload['reply_markup'] = _convert_markup(reply_markup)
+    if allow_sending_without_reply is not None:
+        payload['allow_sending_without_reply'] = allow_sending_without_reply
+    if timeout:
+        payload['connect-timeout'] = timeout
+    return _make_request(token, method_url, params=payload)
+
+
 def send_dice(
         token, chat_id,
         emoji=None, disable_notification=None, reply_to_message_id=None,
@@ -706,7 +752,9 @@ def get_method_by_type(data_type):
 def kick_chat_member(token, chat_id, user_id, until_date=None):
     method_url = 'kickChatMember'
     payload = {'chat_id': chat_id, 'user_id': user_id}
-    if until_date:
+    if isinstance(until_date, datetime):
+        payload['until_date'] = until_date.timestamp()
+    else:
         payload['until_date'] = until_date
     return _make_request(token, method_url, params=payload, method='post')
 
@@ -727,8 +775,6 @@ def restrict_chat_member(
         can_invite_users=None, can_pin_messages=None):
     method_url = 'restrictChatMember'
     permissions = {}
-    if until_date is not None:
-        permissions['until_date'] = until_date
     if can_send_messages is not None:
         permissions['can_send_messages'] = can_send_messages
     if can_send_media_messages is not None:
@@ -747,6 +793,11 @@ def restrict_chat_member(
         permissions['can_pin_messages'] = can_pin_messages
     permissions_json = json.dumps(permissions)
     payload = {'chat_id': chat_id, 'user_id': user_id, 'permissions': permissions_json}
+    if until_date is not None:
+        if isinstance(until_date, datetime):
+            payload['until_date'] = until_date.timestamp()
+        else:
+            payload['until_date'] = until_date
     return _make_request(token, method_url, params=payload, method='post')
 
 
@@ -1243,7 +1294,10 @@ def send_poll(
     if open_period is not None:
         payload['open_period'] = open_period
     if close_date is not None:
-        payload['close_date'] = close_date
+        if isinstance(close_date, datetime):
+            payload['close_date'] = close_date.timestamp()
+        else:
+            payload['close_date'] = close_date
     if is_closed is not None:
         payload['is_closed'] = is_closed
 
@@ -1282,6 +1336,17 @@ def _convert_markup(markup):
     return markup
 
 
+def _convert_entites(entites):
+    if entites is None:
+        return None
+    elif len(entites) == 0:
+        return []
+    elif isinstance(entites[0], types.JsonSerializable):
+        return [entity.to_json() for entity in entites]
+    else:
+        return entites
+
+
 def convert_input_media(media):
     if isinstance(media, types.InputMedia):
         return media.convert_input_media()
